@@ -1,11 +1,12 @@
 {-# LANGUAGE CPP #-}
 #if __GLASGOW_HASKELL__ >= 702
 {-# LANGUAGE Safe #-}
+{-# LANGUAGE DeriveGeneric #-}
 #endif
-#if __GLASGOW_HASKELL__ >= 704
+#if __GLASGOW_HASKELL__ >= 706
 {-# LANGUAGE PolyKinds #-}
 #endif
-#if __GLASGOW_HASKELL__ >= 710
+#if __GLASGOW_HASKELL__ >= 710 && __GLASGOW_HASKELL__ < 802
 {-# LANGUAGE AutoDeriveTypeable #-}
 #endif
 -----------------------------------------------------------------------------
@@ -52,15 +53,21 @@ import Control.Monad.IO.Class
 import Control.Monad.Trans.Class
 import Data.Functor.Identity
 
+#if !(MIN_VERSION_base(4,8,0))
 import Control.Applicative
+#endif
 #if MIN_VERSION_base(4,9,0)
 import qualified Control.Monad.Fail as Fail
 #endif
+#if __GLASGOW_HASKELL__ >= 702
+import GHC.Generics
+#endif
 
 {- |
 Continuation monad.
-@Cont r a@ is a CPS computation that produces an intermediate result
-of type @a@ within a CPS computation whose final result type is @r@.
+@Cont r a@ is a CPS ("continuation-passing style") computation that produces an
+intermediate result of type @a@ within a CPS computation whose final result type
+is @r@.
 
 The @return@ function simply creates a continuation which passes the value on.
 
@@ -72,6 +79,7 @@ type Cont r = ContT r Identity
 -- (The inverse of 'runCont')
 cont :: ((a -> r) -> r) -> Cont r a
 cont f = ContT (\ c -> Identity (f (runIdentity . c)))
+{-# INLINE cont #-}
 
 -- | The result of running a CPS computation with a given final continuation.
 -- (The inverse of 'cont')
@@ -81,6 +89,7 @@ runCont
                         -- the final result (often 'id').
     -> r
 runCont m k = runIdentity (runContT m (Identity . k))
+{-# INLINE runCont #-}
 
 -- | The result of running a CPS computation with the identity as the
 -- final continuation.
@@ -88,6 +97,7 @@ runCont m k = runIdentity (runContT m (Identity . k))
 -- * @'evalCont' ('return' x) = x@
 evalCont :: Cont r r -> r
 evalCont m = runIdentity (evalContT m)
+{-# INLINE evalCont #-}
 
 -- | Apply a function to transform the result of a continuation-passing
 -- computation.
@@ -95,6 +105,7 @@ evalCont m = runIdentity (evalContT m)
 -- * @'runCont' ('mapCont' f m) = f . 'runCont' m@
 mapCont :: (r -> r) -> Cont r a -> Cont r a
 mapCont f = mapContT (Identity . f . runIdentity)
+{-# INLINE mapCont #-}
 
 -- | Apply a function to transform the continuation passed to a CPS
 -- computation.
@@ -102,6 +113,7 @@ mapCont f = mapContT (Identity . f . runIdentity)
 -- * @'runCont' ('withCont' f m) = 'runCont' m . f@
 withCont :: ((b -> r) -> (a -> r)) -> Cont r a -> Cont r b
 withCont f = withContT ((Identity .) . f . (runIdentity .))
+{-# INLINE withCont #-}
 
 -- | @'reset' m@ delimits the continuation of any 'shift' inside @m@.
 --
@@ -109,6 +121,7 @@ withCont f = withContT ((Identity .) . f . (runIdentity .))
 --
 reset :: Cont r r -> Cont r' r
 reset = resetT
+{-# INLINE reset #-}
 
 -- | @'shift' f@ captures the continuation up to the nearest enclosing
 -- 'reset' and passes it to @f@:
@@ -117,10 +130,19 @@ reset = resetT
 --
 shift :: ((a -> r) -> Cont r r) -> Cont r a
 shift f = shiftT (f . (runIdentity .))
+{-# INLINE shift #-}
 
 -- | The continuation monad transformer.
--- Can be used to add continuation handling to other monads.
+-- Can be used to add continuation handling to any type constructor:
+-- the 'Monad' instance and most of the operations do not require @m@
+-- to be a monad.
+--
+-- 'ContT' is not a functor on the category of monads, and many operations
+-- cannot be lifted through it.
 newtype ContT r m a = ContT { runContT :: (a -> m r) -> m r }
+#if __GLASGOW_HASKELL__ >= 702
+    deriving (Generic)
+#endif
 
 -- | The result of running a CPS computation with 'return' as the
 -- final continuation.
@@ -128,13 +150,17 @@ newtype ContT r m a = ContT { runContT :: (a -> m r) -> m r }
 -- * @'evalContT' ('lift' m) = m@
 evalContT :: (Monad m) => ContT r m r -> m r
 evalContT m = runContT m return
+{-# INLINE evalContT #-}
 
 -- | Apply a function to transform the result of a continuation-passing
--- computation.
+-- computation.  This has a more restricted type than the @map@ operations
+-- for other monad transformers, because 'ContT' does not define a functor
+-- in the category of monads.
 --
 -- * @'runContT' ('mapContT' f m) = f . 'runContT' m@
 mapContT :: (m r -> m r) -> ContT r m a -> ContT r m a
 mapContT f m = ContT $ f . runContT m
+{-# INLINE mapContT #-}
 
 -- | Apply a function to transform the continuation passed to a CPS
 -- computation.
@@ -142,30 +168,41 @@ mapContT f m = ContT $ f . runContT m
 -- * @'runContT' ('withContT' f m) = 'runContT' m . f@
 withContT :: ((b -> m r) -> (a -> m r)) -> ContT r m a -> ContT r m b
 withContT f m = ContT $ runContT m . f
+{-# INLINE withContT #-}
 
 instance Functor (ContT r m) where
     fmap f m = ContT $ \ c -> runContT m (c . f)
+    {-# INLINE fmap #-}
 
 instance Applicative (ContT r m) where
     pure x  = ContT ($ x)
+    {-# INLINE pure #-}
     f <*> v = ContT $ \ c -> runContT f $ \ g -> runContT v (c . g)
+    {-# INLINE (<*>) #-}
+    m *> k = m >>= \_ -> k
+    {-# INLINE (*>) #-}
 
 instance Monad (ContT r m) where
 #if !(MIN_VERSION_base(4,8,0))
     return x = ContT ($ x)
+    {-# INLINE return #-}
 #endif
     m >>= k  = ContT $ \ c -> runContT m (\ x -> runContT (k x) c)
+    {-# INLINE (>>=) #-}
 
 #if MIN_VERSION_base(4,9,0)
 instance (Fail.MonadFail m) => Fail.MonadFail (ContT r m) where
     fail msg = ContT $ \ _ -> Fail.fail msg
+    {-# INLINE fail #-}
 #endif
 
 instance MonadTrans (ContT r) where
     lift m = ContT (m >>=)
+    {-# INLINE lift #-}
 
 instance (MonadIO m) => MonadIO (ContT r m) where
     liftIO = lift . liftIO
+    {-# INLINE liftIO #-}
 
 -- | @callCC@ (call-with-current-continuation) calls its argument
 -- function, passing it the current continuation.  It provides
@@ -184,6 +221,7 @@ instance (MonadIO m) => MonadIO (ContT r m) where
 -- layers deep within nested computations.
 callCC :: ((a -> ContT r m b) -> ContT r m a) -> ContT r m a
 callCC f = ContT $ \ c -> runContT (f (\ x -> ContT $ \ _ -> c x)) c
+{-# INLINE callCC #-}
 
 -- | @'resetT' m@ delimits the continuation of any 'shiftT' inside @m@.
 --
@@ -191,6 +229,7 @@ callCC f = ContT $ \ c -> runContT (f (\ x -> ContT $ \ _ -> c x)) c
 --
 resetT :: (Monad m) => ContT r m r -> ContT r' m r
 resetT = lift . evalContT
+{-# INLINE resetT #-}
 
 -- | @'shiftT' f@ captures the continuation up to the nearest enclosing
 -- 'resetT' and passes it to @f@:
@@ -199,6 +238,7 @@ resetT = lift . evalContT
 --
 shiftT :: (Monad m) => ((a -> m r) -> ContT r m r) -> ContT r m a
 shiftT f = ContT (evalContT . f)
+{-# INLINE shiftT #-}
 
 -- | @'liftLocal' ask local@ yields a @local@ function for @'ContT' r m@.
 liftLocal :: (Monad m) => m r' -> ((r' -> r') -> m r -> m r) ->
@@ -206,3 +246,4 @@ liftLocal :: (Monad m) => m r' -> ((r' -> r') -> m r -> m r) ->
 liftLocal ask local f m = ContT $ \ c -> do
     r <- ask
     local f (runContT m (local (const r) . c))
+{-# INLINE liftLocal #-}
